__all__ = ["configure_stdlib_log_forwarding"]

import logging
import logging.config
from collections.abc import Mapping

import structlog
from structlog.typing import FilteringBoundLogger

from ._builtin_parsers import builtin_parsers
from ._parse import Parser, ResolvingParser, default_fallback_parser


def configure_stdlib_log_forwarding(
    *,
    parsers: Mapping[str, Parser] | None = None,
    fallback_parser: Parser | None = None,
    logger: FilteringBoundLogger | None = None,
    validate_structlog_config: bool | None = None,
) -> None:
    if parsers is None:
        parsers = builtin_parsers()
    if fallback_parser is None:
        fallback_parser = default_fallback_parser
    if logger is None:
        logger = structlog.get_logger()

    if validate_structlog_config is None:
        validate_structlog_config = structlog.is_configured()
    if validate_structlog_config:
        if not structlog.is_configured():
            raise RuntimeError(
                "unable to validate structlog for usage with l2sl, because it is not configured"
            )

        config = structlog.get_config()
        if isinstance(
            logger_factory := config.get("logger_factory"),
            structlog.stdlib.LoggerFactory,
        ):
            raise RuntimeError(
                f"l2sl is not compatible with structlog's standard library logging, "
                f"but {logger_factory=} is configured"
            )

    logging.config.dictConfig(
        {
            "version": 1,
            "disable_existing_loggers": False,
            "handlers": {
                "l2sl": {
                    "()": _RecordForwarder,
                    "parser": ResolvingParser(
                        parsers=parsers, fallback=fallback_parser, logger=logger
                    ),
                    "logger": logger,
                }
            },
            "loggers": {
                "root": {"level": "NOTSET", "handlers": ["l2sl"]},
            },
        }
    )


class _RecordForwarder(logging.Handler):
    def __init__(self, parser: Parser, logger: FilteringBoundLogger) -> None:
        super().__init__()
        self._parser = parser
        self._logger = logger

    def emit(self, record: logging.LogRecord) -> None:
        try:
            event_dict = self._parser(record)
        except structlog.exceptions.DropEvent:
            return

        self._logger.log(record.levelno, event_dict.pop("event", ""), **event_dict)



import logging

from structlog.typing import EventDict

from .._parse import RegexpEventParser
from . import register_builtin_parser

tornado_access = register_builtin_parser(RegexpEventParser(), logger="tornado.access")


@tornado_access.register_event_handler(
    r"(?P<status_code>\d{3}) (?P<method>[A-Z]+) (?P<endpoint>.*) \((?P<origin>.*)\) (?P<elapsed_time>\d+\.\d+)ms"
)
def event_handler(groups: dict[str, str], record: logging.LogRecord) -> EventDict:
    return {
        "event": "request",
        "elapsed_time": float(groups.pop("elapsed_time")) * 1e-3,
    } | groups



import logging

from structlog.typing import EventDict

from .._parse import RegexpEventParser
from . import register_builtin_parser

panel_io = register_builtin_parser(RegexpEventParser(), logger="panel.io")


@panel_io.register_event_handler(r"(?P<event>Session) (?P<id>\d+) (?P<state>.+)")
def session(groups: dict[str, str], record: logging.LogRecord) -> EventDict:
    return groups


panel_viewable = register_builtin_parser(RegexpEventParser(), logger="panel.viewable")


@panel_viewable.register_event_handler(
    r"Session \d+ (?P<event>(received|finished processing) events)"
)
def handler(groups: dict[str, str], record: logging.LogRecord) -> EventDict:
    assert record.args is not None
    session_id, events = record.args
    return {"event": groups["event"], "sesion_id": session_id, "events": events}



from __future__ import annotations

__all__ = ["LogLevel", "LogLevelNumber", "StructlogLogLevelName", "StdlibLogLevelName"]

import functools
from typing import Any, ClassVar, Literal, cast

try:
    import pydantic
    from pydantic_core import core_schema

    _PYDANTIC_2_AVAILABLE = tuple(map(int, pydantic.__version__.split(".")[:3])) >= (2,)
except (ImportError, AttributeError):
    _PYDANTIC_2_AVAILABLE = False

LogLevelNumber = Literal[0, 10, 20, 30, 40, 50]
StdlibLogLevelName = Literal["NOTSET", "DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"]
StructlogLogLevelName = Literal[
    "notset", "debug", "info", "warning", "warn", "error", "exception", "critical"
]


@functools.total_ordering
class LogLevel:
    _NUMBER_TO_STDLIB_NAME: ClassVar[dict[LogLevelNumber, StdlibLogLevelName]] = {
        0: "NOTSET",
        10: "DEBUG",
        20: "INFO",
        30: "WARNING",
        40: "ERROR",
        50: "CRITICAL",
    }
    _STRUCTLOG_NAME_TO_STDLIB_NAME: dict[StructlogLogLevelName, StdlibLogLevelName] = {
        "notset": "NOTSET",
        "debug": "DEBUG",
        "info": "INFO",
        "warning": "WARNING",
        "warn": "WARNING",
        "error": "ERROR",
        "exception": "ERROR",
        "critical": "CRITICAL",
    }
    _STDLIB_NAME_TO_NUMBER: ClassVar[dict[StdlibLogLevelName, LogLevelNumber]] = {
        v: k for k, v in _NUMBER_TO_STDLIB_NAME.items()
    }
    _STDLIB_NAME_TO_STRUCTLOG_NAME: dict[StdlibLogLevelName, StructlogLogLevelName] = {
        "NOTSET": "notset",
        "DEBUG": "debug",
        "INFO": "info",
        "WARNING": "warning",
        "ERROR": "error",
        "CRITICAL": "critical",
    }

    def __init__(
        self, level: LogLevelNumber | StdlibLogLevelName | StructlogLogLevelName
    ) -> None:
        if (number := cast(LogLevelNumber, level)) in self._NUMBER_TO_STDLIB_NAME:
            stdlib_name = self._NUMBER_TO_STDLIB_NAME[number]
            structlog_name = self._STDLIB_NAME_TO_STRUCTLOG_NAME[stdlib_name]
        elif (
            stdlib_name := cast(StdlibLogLevelName, level)
        ) in self._STDLIB_NAME_TO_NUMBER:
            number = self._STDLIB_NAME_TO_NUMBER[stdlib_name]
            structlog_name = self._STDLIB_NAME_TO_STRUCTLOG_NAME[stdlib_name]
        elif (
            structlog_name := cast(StructlogLogLevelName, level)
        ) in self._STRUCTLOG_NAME_TO_STDLIB_NAME:
            stdlib_name = self._STRUCTLOG_NAME_TO_STDLIB_NAME[structlog_name]
            number = self._STDLIB_NAME_TO_NUMBER[stdlib_name]
        else:
            raise ValueError

        self._number = number
        self._stdlib_name = stdlib_name
        self._structlog_name = structlog_name

    @property
    def number(self) -> LogLevelNumber:
        return self._number

    @property
    def stdlib_name(self) -> StdlibLogLevelName:
        return self._stdlib_name

    @property
    def structlog_name(self) -> StructlogLogLevelName:
        return self._structlog_name

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, LogLevel):
            try:
                other = LogLevel(other)
            except ValueError:
                return NotImplemented
        other: LogLevel

        return self.number == other.number

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, LogLevel):
            try:
                other = LogLevel(other)
            except ValueError:
                return NotImplemented
        other: LogLevel

        return self.number < other.number

    def __str__(self) -> str:
        return self.stdlib_name

    def __repr__(self) -> str:
        return f"{type(self)}(number={self.number}, stdlib_name={self.stdlib_name}, structlog_name={self.structlog_name})"

    def __int__(self) -> int:
        return self.number

    if _PYDANTIC_2_AVAILABLE:

        @classmethod
        def __get_pydantic_core_schema__(
            cls, _source_type: Any, _handler: pydantic.GetCoreSchemaHandler
        ) -> core_schema.CoreSchema:
            json_schema = core_schema.chain_schema(
                [
                    core_schema.literal_schema(
                        [
                            *cls._NUMBER_TO_STDLIB_NAME,
                            *cls._STDLIB_NAME_TO_NUMBER,
                            *cls._STRUCTLOG_NAME_TO_STDLIB_NAME,
                        ]
                    ),
                    core_schema.no_info_plain_validator_function(cls),
                ]
            )

            return core_schema.json_or_python_schema(
                json_schema=json_schema,
                python_schema=core_schema.union_schema(
                    [
                        core_schema.is_instance_schema(cls),
                        json_schema,
                    ]
                ),
                serialization=core_schema.plain_serializer_function_ser_schema(str),
            )



import logging

from structlog.typing import EventDict

from .._parse import RegexpEventParser
from . import register_builtin_parser

uvicorn_error = register_builtin_parser(RegexpEventParser(), logger="uvicorn.error")


@uvicorn_error.register_event_handler(r"(?P<event>(Started|Finished) server process)")
def server_process(groups: dict[str, str], record: logging.LogRecord) -> EventDict:
    assert record.args is not None
    (pid,) = record.args
    return groups | {"pid": pid}


@uvicorn_error.register_event_handler(r"(?P<event>Uvicorn running) on")
def uvicorn_running(groups: dict[str, str], record: logging.LogRecord) -> EventDict:
    assert record.args is not None
    _, host, port = record.args
    return groups | {"host": host, "port": port}


@register_builtin_parser(logger="uvicorn.access")
def uvicorn_access(record: logging.LogRecord) -> EventDict:
    assert record.args is not None
    origin, method, endpoint, protocol_version, status_code = record.args
    return {
        "event": "request",
        "origin": origin,
        "method": method,
        "endpoint": endpoint,
        "protocol": f"HTTP/{protocol_version}",
        "status_code": status_code,
    }



import json
import operator

import pydantic
import pytest

from l2sl import LogLevel


class TestLogLevel:
    @pytest.mark.parametrize(
        ("type", "value"),
        [
            *[("number", v) for v in LogLevel._NUMBER_TO_STDLIB_NAME],
            *[("stdlib_name", v) for v in LogLevel._STDLIB_NAME_TO_NUMBER],
            *[("structlog_name", v) for v in LogLevel._STRUCTLOG_NAME_TO_STDLIB_NAME],
        ],
    )
    def test_parsing(self, type, value):
        log_level = LogLevel(value)
        assert getattr(log_level, type) == value

    @pytest.mark.parametrize(
        ("a", "op", "b"),
        [
            ("info", operator.lt, "warning"),
            ("info", operator.lt, "WARNING"),
            ("info", operator.lt, 30),
            ("info", operator.eq, "info"),
            ("info", operator.eq, "INFO"),
            ("info", operator.eq, 20),
            ("info", operator.gt, "debug"),
            ("info", operator.gt, "DEBUG"),
            ("info", operator.gt, 10),
        ],
    )
    @pytest.mark.parametrize("as_log_level", ["left", "right", "both"])
    def test_comparison(self, a, op, b, as_log_level):
        match as_log_level:
            case "left":
                a = LogLevel(a)
            case "right":
                b = LogLevel(b)
            case "both":
                a = LogLevel(a)
                b = LogLevel(b)
            case _:
                raise ValueError(f"{as_log_level=}")

        assert op(a, b)

    @pytest.mark.parametrize("level", ["info", "INFO", 20])
    def test_str(self, level):
        log_level = LogLevel(level)
        assert str(log_level) == log_level.stdlib_name

    @pytest.mark.parametrize("level", ["info", "INFO", 20])
    def test_repr_smoke(self, level):
        repr(level)

    @pytest.mark.parametrize("level", ["info", "INFO", 20, LogLevel("info")])
    def test_pydantic_validate_python(self, level):
        log_level = level if isinstance(level, LogLevel) else LogLevel(level)

        ta = pydantic.TypeAdapter(LogLevel)
        assert ta.validate_python(level) == log_level

    def test_pydantic_dump_python(self):
        log_level = LogLevel("info")

        ta = pydantic.TypeAdapter(LogLevel)
        assert ta.dump_python(log_level) == str(log_level)

    @pytest.mark.parametrize("level", ["info", "INFO", 20])
    def test_int(self, level):
        log_level = LogLevel(level)
        assert int(log_level) == log_level.number

    @pytest.mark.parametrize("level", ["info", "INFO", 20])
    def test_pydantic_validate_json(self, level):
        log_level = LogLevel(level)

        ta = pydantic.TypeAdapter(LogLevel)
        assert ta.validate_json(json.dumps(level)) == log_level

    def test_pydantic_dump_json(self):
        log_level = LogLevel("info")

        ta = pydantic.TypeAdapter(LogLevel)
        assert ta.dump_json(log_level) == json.dumps(str(log_level)).encode()



__all__ = ["builtin_parsers", "register_builtin_parser"]

import importlib
from pathlib import Path
from typing import Callable, TypeVar, overload

from .._parse import Parser

_BUILTIN: dict[str, Parser] = {}

TParser = TypeVar("TParser", bound=Parser)


@overload
def register_builtin_parser(parser: TParser, /, *, logger: str) -> TParser: ...


@overload
def register_builtin_parser(
    parser: None = None, /, *, logger: str
) -> Callable[[TParser], TParser]: ...


def register_builtin_parser(
    parser: TParser | None = None, /, *, logger: str
) -> TParser | Callable[[TParser], TParser]:
    def register(parser: TParser) -> TParser:
        _BUILTIN[logger] = parser
        return parser

    if parser is None:
        return register
    else:
        return register(parser)


def builtin_parsers() -> dict[str, Parser]:
    if not _BUILTIN:
        for p in sorted(Path(__file__).parent.glob("[!_]*.py")):
            importlib.import_module(f"{__package__}.{p.stem}")
    return _BUILTIN.copy()



import pytest
import structlog

import l2sl


def test_configure_stdlib_log_forwarding_unable_to_validate():
    with pytest.raises(RuntimeError, match="not configured"):
        l2sl.configure_stdlib_log_forwarding(validate_structlog_config=True)


def test_configure_stdlib_log_forwarding_not_compatible():
    structlog.configure(logger_factory=structlog.stdlib.LoggerFactory())

    with pytest.raises(RuntimeError, match="not compatible"):
        l2sl.configure_stdlib_log_forwarding()



import pytest

from l2sl._parse import _LoggerResolver


@pytest.mark.parametrize(
    ("available_loggers", "logger", "expected"),
    [
        (["foo", "bar"], "foo", "foo"),
        (["foo", "bar"], "baz", None),
        (["foo.boo", "bar"], "foo", None),
        (["foo.boo", "bar"], "foo.boo", "foo.boo"),
        (["foo", "bar"], "foo.boo", "foo"),
    ],
)
def test_logger_resolver(available_loggers, logger, expected):
    assert expected in available_loggers or expected is None

    logger_resolver = _LoggerResolver(available_loggers)
    assert logger_resolver(logger) == expected



try:
    from ._version import __version__
except ModuleNotFoundError:
    import warnings

    warnings.warn("l2sl was not properly installed!")
    del warnings

    __version__ = "UNKNOWN"

from ._builtin_parsers import builtin_parsers
from ._forward import configure_stdlib_log_forwarding
from ._log_level import (
    LogLevel,
    LogLevelNumber,
    StdlibLogLevelName,
    StructlogLogLevelName,
)
from ._parse import (
    Parser,
    RegexpEventHandler,
    RegexpEventParser,
    default_fallback_parser,
)

__all__ = [
    "LogLevel",
    "LogLevelNumber",
    "StdlibLogLevelName",
    "StructlogLogLevelName",
    "__version__",
    "configure_stdlib_log_forwarding",
    "Parser",
    "builtin_parsers",
    "default_fallback_parser",
    "RegexpEventHandler",
    "RegexpEventParser",
]



import logging

from structlog.typing import EventDict

from . import register_builtin_parser


@register_builtin_parser(logger="httpx")
def httpx(record: logging.LogRecord) -> EventDict:
    assert record.args is not None
    method, url, protocol, status_code, _ = record.args
    return {
        "event": "request",
        "method": method,
        "url": str(url),
        "protocol": protocol,
        "status_code": status_code,
    }



import logging

from structlog.typing import EventDict

from .._parse import RegexpEventParser
from . import register_builtin_parser

bokeh_server_server = register_builtin_parser(
    RegexpEventParser(), logger="bokeh.server.server"
)


@bokeh_server_server.register_event_handler(
    r"(?P<event>Starting Bokeh server) version (?P<bokeh_version>\d+\.\d+\.\d+) \(running on Tornado (?P<tornado_version>\d+\.\d+\.\d+)\)"
)
def starting_server(groups: dict[str, str], record: logging.LogRecord) -> EventDict:
    return groups


bokeh_server_tornado = register_builtin_parser(
    RegexpEventParser(), logger="bokeh.server.tornado"
)


@bokeh_server_tornado.register_event_handler(r"\[pid \d+\] \d+ clients connected")
def clients(groups: dict[str, str], record: logging.LogRecord) -> EventDict:
    assert record.args is not None
    pid, number = record.args
    return {"event": "clients", "pid": pid, "number": number}


@bokeh_server_tornado.register_event_handler(
    r"\[pid \d+\]\s+.*? has \d+ sessions with \d+ unused"
)
def sessions(groups: dict[str, str], record: logging.LogRecord) -> EventDict:
    assert record.args is not None
    pid, endpoint, number, unused = record.args
    return {
        "event": "sessions",
        "pid": pid,
        "endpoint": endpoint,
        "number": number,
        "unused": unused,
    }



from __future__ import annotations

__all__ = [
    "ResolvingParser",
    "Parser",
    "default_fallback_parser",
    "RegexpEventParser",
    "RegexpEventHandler",
]

import functools
import logging
import re
import secrets
import sys
import uuid
from collections.abc import Iterable, Mapping
from typing import Any, Callable

import structlog
from structlog.typing import EventDict, FilteringBoundLogger

Parser = Callable[[logging.LogRecord], EventDict]

_DEFAULT_RECORD_ATTRIBUTES = set(
    logging.LogRecord("", logging.DEBUG, "", 0, "", None, None).__dict__.keys()
)


def _extract_record_extra(record: logging.LogRecord) -> dict[str, Any]:
    return {
        k: v for k, v in record.__dict__.items() if k not in _DEFAULT_RECORD_ATTRIBUTES
    }


def default_fallback_parser(record: logging.LogRecord) -> EventDict:
    event_dict: dict[str, Any] = _extract_record_extra(record) | {
        "event": record.getMessage()
    }
    if exc_info := record.exc_info:
        if isinstance(exc_info, BaseException):
            exc_info = (type(exc_info), exc_info, exc_info.__traceback__)
        elif not isinstance(exc_info, tuple):
            exc_info = sys.exc_info()
        event_dict["exc_info"] = exc_info

    if stack_info := record.stack_info:
        event_dict["stack_info"] = stack_info

    return event_dict


class ResolvingParser:
    def __init__(
        self,
        *,
        parsers: Mapping[str, Parser],
        fallback: Parser = default_fallback_parser,
        logger: FilteringBoundLogger,
    ) -> None:
        self._parsers = parsers
        self._fallback = fallback
        self._logger = logger

        self._logger_resolver = _LoggerResolver(self._parsers.keys())

    def __call__(self, record: logging.LogRecord) -> EventDict:
        logger = record.name
        resolved_logger = self._logger_resolver(logger)
        parser = (
            self._parsers[resolved_logger]
            if resolved_logger is not None
            else self._fallback
        )

        try:
            event_dict = parser(record)
        except structlog.exceptions.DropEvent:
            raise
        except Exception as exc_info:
            event_dict = self._fallback(record)
            l2sl_parser_error_id = event_dict["l2sl_parser_error_id"] = str(
                uuid.uuid4()
            )
            self._logger.error(
                "failed to parse record",
                logger="l2sl",
                exc_info=exc_info,
                l2sl_parser_error_id=l2sl_parser_error_id,
            )

        event_dict["logger"] = logger
        return event_dict


class _LoggerResolver:
    def __init__(self, available_loggers: Iterable[str]) -> None:
        self._available_loggers = [l.split(".") for l in available_loggers]

    @functools.lru_cache()
    def __call__(self, logger: str) -> str | None:
        l = logger.split(".")
        applicable_loggers = sorted(
            (
                a
                for a in self._available_loggers
                if len(l) >= len(a) and l[: len(a)] == a
            ),
            key=len,
        )
        if not applicable_loggers:
            return None

        return ".".join(applicable_loggers[-1])


RegexpEventHandler = Callable[[dict[str, str], logging.LogRecord], EventDict]


class RegexpEventParser:
    def __init__(self, fallback: Parser = default_fallback_parser) -> None:
        self._event_handlers: dict[str, tuple[str, RegexpEventHandler]] = {}
        self._fallback = fallback

    def register_event_handler(
        self, pattern: str
    ) -> Callable[[RegexpEventHandler], RegexpEventHandler]:
        def decorator(eh: RegexpEventHandler) -> RegexpEventHandler:
            self._event_handlers[_unique_regex_identifier()] = (pattern, eh)
            return eh

        return decorator

    @functools.cached_property
    def _parser(self) -> _RegexpEventParser:
        return _RegexpEventParser(self._event_handlers, self._fallback)

    def __call__(self, record: logging.LogRecord) -> EventDict:
        return self._parser(record)


class _RegexpEventParser:
    def __init__(
        self,
        event_handlers: dict[str, tuple[str, RegexpEventHandler]],
        fallback: Parser,
    ) -> None:
        self._pattern, self._event_map = self._compile(event_handlers)
        self._fallback = fallback

    _GROUP_PATTERN = re.compile(r"\(\?P<(?P<group>\w+)>")

    def _compile(
        self, event_handlers: dict[str, tuple[str, RegexpEventHandler]]
    ) -> tuple[re.Pattern[str], dict[str, tuple[dict[str, str], RegexpEventHandler]]]:
        event_patterns: dict[str, str] = {}
        event_map = {}
        for event_id, (event_pattern, event_handler) in event_handlers.items():
            group_map = {
                group: _unique_regex_identifier()
                for group in self._GROUP_PATTERN.findall(event_pattern)
            }

            event_patterns[event_id] = self._GROUP_PATTERN.sub(
                lambda match: f"(?P<{group_map[match['group']]}>", event_pattern
            )
            event_map[event_id] = (group_map, event_handler)

        pattern = "|".join(
            f"(?P<{event_id}>{pattern})" for event_id, pattern in event_patterns.items()
        )
        pattern = f"({pattern})"

        return re.compile(pattern), event_map

    def __call__(self, record: logging.LogRecord) -> EventDict:
        event = record.getMessage()
        match = self._pattern.match(event)
        if not match:
            return self._fallback(record)

        groups = match.groupdict()
        group_map, event_handler = next(
            v for id, v in self._event_map.items() if groups[id] is not None
        )

        return event_handler(
            {group: groups[group_id] for group, group_id in group_map.items()}, record
        )


def _unique_regex_identifier() -> str:
    return f"_{secrets.token_hex(8)}"
